use super::{archive, package};
use crate::{db, error::Result};

use std::{
    collections::HashMap,
    path::{Path, PathBuf},
    sync::{
        atomic::{AtomicBool, AtomicU32, Ordering},
        Arc,
    },
};

use futures::StreamExt;
use sea_orm::{
    ActiveModelTrait, ColumnTrait, DbConn, EntityTrait, ModelTrait, QueryFilter, QuerySelect,
    Related, RelationTrait, Set, TransactionTrait,
};
use sea_query::{Expr, Query};
use tokio::{
    io::AsyncRead,
    sync::{Mutex, Notify, RwLock, Semaphore},
};
use uuid::Uuid;

pub const ANY_ARCH: &'static str = "any";
pub const REPOS_DIR: &'static str = "repos";

#[derive(Default)]
pub struct RepoState {
    queued_pkgs: AtomicU32,
    sync_queued: AtomicBool,
    sync_notify: Notify,
}

pub struct DistroMgr {
    distro_dir: PathBuf,
    distro_id: i32,
    conn: DbConn,
    repos: RwLock<HashMap<i32, Arc<RepoState>>>,
    sync_lock: Mutex<()>,
    pkg_sema: Semaphore,
}

impl DistroMgr {
    pub async fn new<P: AsRef<Path>>(distro_dir: P, distro_id: i32, conn: DbConn) -> Result<Self> {
        if !tokio::fs::try_exists(&distro_dir).await? {
            tokio::fs::create_dir(&distro_dir).await?;
        }

        let repos_dir = distro_dir.as_ref().join(REPOS_DIR);

        if !tokio::fs::try_exists(&repos_dir).await? {
            tokio::fs::create_dir(repos_dir).await?;
        }

        Ok(Self {
            distro_dir: distro_dir.as_ref().to_path_buf(),
            distro_id,
            conn,
            sync_lock: Mutex::new(()),
            pkg_sema: Semaphore::new(1),
            repos: RwLock::new(HashMap::new()),
        })
    }

    pub async fn schedule_sync(&self, repo_id: i32) -> Result<()> {
        let state = {
            let repos = self.repos.read().await;
            repos.get(&repo_id).map(Arc::clone)
        };

        if state.is_none() {
            tracing::debug!("is none");
            return Ok(());
        }

        let state = state.unwrap();

        let res =
            state
                .sync_queued
                .compare_exchange(false, true, Ordering::SeqCst, Ordering::SeqCst);

        // Already a sync job scheduled, so this one can simply quit
        if res.is_err() {
            tracing::debug!("shit");
            return Ok(());
        }

        // If the queue is not empty, we wait for a notification that it is before syncing
        if state.queued_pkgs.load(Ordering::SeqCst) > 0 {
            tracing::debug!("sync waiter waiting");
            state.sync_notify.notified().await;
            tracing::debug!("sync waiter notified");
        }

        self.sync_repo(repo_id).await
    }

    /// Generate archive databases for all known architectures in the repository, including the
    /// "any" architecture.
    pub async fn sync_repo(&self, repo_id: i32) -> Result<()> {
        let _guard = self.sync_lock.lock().await;

        let repo = crate::db::query::repo::by_id(&self.conn, repo_id).await?;

        if repo.is_none() {
            return Ok(());
        }

        let repo = repo.unwrap();

        let mut archs = repo
            .find_related(db::Package)
            .select_only()
            .column(db::package::Column::Arch)
            .distinct()
            .into_tuple::<String>()
            .stream(&self.conn)
            .await?;

        while let Some(arch) = archs.next().await.transpose()? {
            self.generate_archives(&repo, &arch).await?;
        }

        Ok(())
    }

    /// Generate the archive databases for the given repository and architecture.
    async fn generate_archives(&self, repo: &db::repo::Model, arch: &str) -> Result<()> {
        let [tmp_ar_db_path, tmp_ar_files_path, files_tmp_file_path, desc_tmp_file_path] =
            self.random_file_paths();
        let ar_db = archive::RepoArchiveWriter::open(&tmp_ar_db_path).await?;
        let ar_files = archive::RepoArchiveWriter::open(&tmp_ar_files_path).await?;

        // Query all packages in the repo that have the given architecture or the "any"
        // architecture
        let mut pkgs = repo
            .find_related(crate::db::Package)
            .filter(db::package::Column::Arch.is_in([arch, ANY_ARCH]))
            .filter(
                db::package::Column::Id.in_subquery(
                    Query::select()
                        .expr(db::package::Column::Id.max())
                        .from(db::package::Entity)
                        .group_by_columns([db::package::Column::Arch, db::package::Column::Name])
                        .to_owned(),
                ),
            )
            .stream(&self.conn)
            .await?;

        let mut commited_ids: Vec<i32> = Vec::new();

        while let Some(pkg) = pkgs.next().await.transpose()? {
            commited_ids.push(pkg.id);

            let mut files_tmp_file = tokio::fs::File::create(&files_tmp_file_path).await?;
            let mut desc_tmp_file = tokio::fs::File::create(&desc_tmp_file_path).await?;

            package::write_files(&self.conn, &mut files_tmp_file, &pkg).await?;
            package::write_desc(&self.conn, &mut desc_tmp_file, &pkg).await?;

            let full_name = format!("{}-{}", pkg.name, pkg.version);

            ar_db
                .add_entry(&full_name, &desc_tmp_file_path, true)
                .await?;
            ar_files
                .add_entry(&full_name, &desc_tmp_file_path, true)
                .await?;
            ar_files
                .add_entry(&full_name, &files_tmp_file_path, false)
                .await?;
        }

        // Cleanup
        ar_db.close().await?;
        ar_files.close().await?;

        let repo_dir = self.distro_dir.join(&repo.name);

        // Move the db archives to their respective places
        tokio::fs::rename(tmp_ar_db_path, repo_dir.join(format!("{}.db.tar.gz", arch))).await?;
        tokio::fs::rename(
            tmp_ar_files_path,
            repo_dir.join(format!("{}.files.tar.gz", arch)),
        )
        .await?;

        // Only after we have successfully written everything to disk do we update the database.
        // This order ensures any failure can be recovered, as the database is our single source of
        // truth.
        db::Package::update_many()
            .col_expr(
                db::package::Column::State,
                Expr::value(db::PackageState::Committed),
            )
            .filter(db::package::Column::Id.is_in(commited_ids))
            .exec(&self.conn)
            .await?;

        // If this fails there's no point in failing the function + if there were no packages in
        // the repo, this fails anyway because the temp file doesn't exist
        let _ = tokio::fs::remove_file(desc_tmp_file_path).await;
        let _ = tokio::fs::remove_file(files_tmp_file_path).await;

        tracing::info!(
            "Package archives generated for '{}' ('{}')",
            &repo.name,
            arch
        );

        Ok(())
    }

    async fn get_or_create_repo(&self, repo: &str) -> Result<db::repo::Model> {
        let mut repos = self.repos.write().await;

        if let Some(repo) = db::query::repo::by_name(&self.conn, repo).await? {
            Ok(repo)
        } else {
            tokio::fs::create_dir(self.distro_dir.join(repo)).await?;
            let repo = db::query::repo::insert(&self.conn, self.distro_id, repo, None).await?;

            repos.insert(repo.id, Arc::new(RepoState::default()));

            Ok(repo)
        }
    }

    /// Remove the repo with the given name, if it existed
    pub async fn remove_repo(&self, repo: &str) -> Result<bool> {
        let res = db::query::repo::by_name(&self.conn, repo).await?;

        if let Some(repo_entry) = res {
            // Remove repository from database
            repo_entry.delete(&self.conn).await?;

            // Remove files from file system
            tokio::fs::remove_dir_all(self.distro_dir.join(repo)).await?;

            Ok(true)
        } else {
            Ok(false)
        }
    }

    /// Remove all packages from the repository with the given arch.
    pub async fn remove_repo_arch(&self, repo: &str, arch: &str) -> Result<bool> {
        let repo = db::query::repo::by_name(&self.conn, repo).await?;

        if let Some(repo) = repo {
            let mut pkgs = repo
                .find_related(db::Package)
                .filter(db::package::Column::Arch.eq(arch))
                .stream(&self.conn)
                .await?;

            while let Some(pkg) = pkgs.next().await.transpose()? {
                let path = self
                    .distro_dir
                    .join(&repo.name)
                    .join(super::package::filename(&pkg));
                tokio::fs::remove_file(path).await?;

                pkg.delete(&self.conn).await?;
            }

            tokio::fs::remove_file(
                self.distro_dir
                    .join(&repo.name)
                    .join(format!("{}.db.tar.gz", arch)),
            )
            .await?;
            tokio::fs::remove_file(
                self.distro_dir
                    .join(&repo.name)
                    .join(format!("{}.files.tar.gz", arch)),
            )
            .await?;

            // If we removed all "any" packages, we need to resync all databases
            if arch == ANY_ARCH {
                //self.sync_repo(&repo.name).await?;
            }

            Ok(true)
        } else {
            Ok(false)
        }
    }

    pub async fn remove_pkg(&self, repo: &str, arch: &str, name: &str) -> Result<bool> {
        let repo = db::query::repo::by_name(&self.conn, repo).await?;

        if let Some(repo) = repo {
            let pkg =
                db::query::package::by_fields(&self.conn, repo.id, arch, name, None, None).await?;

            if let Some(pkg) = pkg {
                // Remove package from database & file system
                tokio::fs::remove_file(
                    self.distro_dir
                        .join(&repo.name)
                        .join(super::package::filename(&pkg)),
                )
                .await?;
                pkg.delete(&self.conn).await?;

                //if arch == ANY_ARCH {
                //    self.sync_repo(&repo.name).await?;
                //} else {
                //    self.generate_archives(&repo.name, arch).await?;
                //}

                Ok(true)
            } else {
                Ok(false)
            }
        } else {
            Ok(false)
        }
    }

    async fn _add_pkg_from_path<P: AsRef<Path>>(
        &self,
        path: P,
        repo: &db::repo::Model,
    ) -> crate::Result<db::package::Model> {
        let path_clone = path.as_ref().to_path_buf();
        let pkg = tokio::task::spawn_blocking(move || package::Package::open(path_clone))
            .await
            .unwrap()?;

        // TODO prevent database from being updated but file failing to move to repo dir?
        let pkg = db::query::package::insert(&self.conn, repo.id, pkg).await?;

        let queue_path = self.distro_dir.join(&repo.name).join(pkg.id.to_string());
        tokio::fs::rename(path.as_ref(), queue_path).await?;

        tracing::info!(
            "Added '{}-{}' to repository '{}' ({})",
            pkg.name,
            pkg.version,
            repo.name,
            pkg.arch
        );

        Ok(pkg)
    }

    pub async fn add_pkg_from_path<P: AsRef<Path>>(
        &self,
        path: P,
        repo: &str,
    ) -> crate::Result<(i32, String, String, String)> {
        let repo = self.get_or_create_repo(repo).await?;

        {
            let repos = self.repos.read().await;

            if let Some(state) = repos.get(&repo.id) {
                state.queued_pkgs.fetch_add(1, Ordering::SeqCst);
            }
        }

        let _guard = self.pkg_sema.acquire().await.unwrap();
        let res = self._add_pkg_from_path(path, &repo).await;

        match res {
            Ok(pkg) => {
                let repos = self.repos.read().await;

                if let Some(state) = repos.get(&repo.id) {
                    let old = state.queued_pkgs.fetch_sub(1, Ordering::SeqCst);

                    if old - 1 == 0 {
                        state.sync_notify.notify_one();
                    }
                }

                Ok((repo.id, pkg.name, pkg.version, pkg.arch))
            }
            Err(e) => Err(e),
        }

        // If the package already exists in the database, we remove it first
        //let res = db::query::package::by_fields(
        //    &self.conn,
        //    repo.id,
        //    &pkg.info.arch,
        //    &pkg.info.name,
        //    None,
        //    None,
        //)
        //.await?;
        //
        //if let Some(entry) = res {
        //    entry.delete(&self.conn).await?;
        //}

        //let dest_pkg_path = repo_dir.join(pkg.file_name());
        //
        //// Insert new package into database
        //let name = pkg.info.name.clone();
        //let version = pkg.info.version.clone();
        //let arch = pkg.info.arch.clone();
        //db::query::package::insert(&self.conn, repo.id, pkg).await?;
        //
        //// Move the package to its final resting place
        //tokio::fs::rename(tmp_file_path, dest_pkg_path).await?;
        //
        // Synchronize archive databases
        //if arch == ANY_ARCH {
        //    self.generate_archives_all(&repo.name).await?;
        //} else {
        //    self.generate_archives(&repo.name, &arch).await?;
        //}
    }

    ///  Generate a path to a unique file that can be used as a temporary file
    pub fn random_file_paths<const C: usize>(&self) -> [PathBuf; C] {
        std::array::from_fn(|_| {
            let uuid: uuid::fmt::Simple = Uuid::new_v4().into();
            self.distro_dir.join(uuid.to_string())
        })
    }
}
